---
author: François Pelletier
title: Notes de cours - examen partiel
fig_caption: yes
highlight: pygments
number_sections: yes
toc: yes
toc_depth: 3
geometry: "left=2cm,right=2cm,top=2cm,bottom=2cm"
fontsize: 12pt
---

# Expressions régulières

## Définitions

- **Patron**: représentation compacte d’un ensemble de chaînes de caractères
- **Recherche**: filtrer les textes qui correspondent au patron

## Types de recherches

- Litéral
- Classe de caractères
  - Alphanumériques: `/\w/`
  - Numériques: `/\d/`
- Registre de valeurs
  - Lettres: `/[a-z]/`
  - Chiffres: `/[0-9]/`
- Négation
  - Aucun chiffre: `/[^0-9]/`
  - Aucun espace: `/[^\s]/`
- Optionnel
  - Orthographes alternatives: `/colou?r/`
  - Mot facultatif: `/(\w+)?/`
- N'importe quel caractère: `/./`
- Alias
- Compteurs
  - Zéro ou plusieurs espaces (en début de ligne): `/^\s*/`
  - Un ou plusieurs caractères (un mot): `/\w+/`
- Groupes de capture
  - Multiple, optionnel ou non-quantifié: `/(groupe 1)+(groupe 2)?(groupe 3)*/`
- Conditions
  - `/(?(?=si_expression)alors|sinon)/`
 
## Erreurs possibles

- Faux positifs
  - n'auraient pas dû correspondre au patron
  - métrique: précision
- Faux négatifs
  - auraient dû correspondre au patron
  - métrique: rappel
  
## Python

Module **re**

|fonction|usage|
|---|---|
|compile|générer une version compilée de l'expression régulière|
|match|rechercher depuis le début d'une chaîne de caractères|
|search|rechercher dans une chaîne de caractères|
|sub|remplacer la première occurence d'un patron|
 
## Automates à états finis

- Servent à implémenter une version compilée des expressions régulières
- Deux types:
  - Déterministes
  - Probabilistes (non-déterministes)
- Représentés sous forme de graphes
- Alphabet: ensemble fini de symboles. Peuvent désigner de plus gros objets.
- Tableau: représente la fonction de transition $q_{i+1} = f(q_i,I)$ pour un état $q_i$ et un symbole en entrée $I$.

### Exemple: langage des moutons

- Expression régulière: `/baa+!/`
- Graphe: 
  - ![FSA sheep language](fsa_sheep.png)
- Caractéristiques:
  - 5 états:
  - Alphabet: a,b,!
  - $q_0$ est l'état de départ
  - $q_4$ est un état terminal
  - 5 transitions
- Tableau

| |b|a|!|e|
|---|---|---|---|---|
|0|1||||
|1||1|||
|2||3|||
|3||3|4||
|4:|||||

### Reconnaissance déterministe

Processus qui:

- Démarre dans l'état initial
- Examine le symbole en entrée
- Consulte la table
- Sélectionne le nouvel état et met à jour le ruban
- Continue jusqu'à ce qu'on atteigne le bout du ruban

Algorithme déterministe: D-Recognise

- Une seule transition possible à chaque instant
- Interpréteur dirigé par une table de transition
- Trivial: On sait qu'on démarre sur le premier caractère et qu'on a un succès si on atteint l'état terminal à la fin du ruban. Mais, on peut atteindre l'état terminal avant la fin du ruban, et débuter ailleurs qu'au début du ruban.

Vérification d'expressions régulières:

- Traduire l'expression en table de transition
- Passer la table et la chaîne à un interpréteur

### Reconnaissance non-déterministe

Deux approches de base:

- Convertir un automate non-déterministe en automate déterministe. 
  - Transition $\epsilon$ qui n'avance pas le ruban.
- Gérer les différentes options comme une exploration dans un espace d'états
  - Il existe au moins un chemin complet pour une chaîne définie par le langage de l'automate.
  - Les chemins ne mènent pas tous à un état final.
  - Aucun chemin ne mène à un état final si la chaîne ne fait pas partie du langage
- Les états de l'espace sont des paires (position du ruban, état de l'automate)
- On garde une trace des états non-explorés. Exploration en profondeur.
- Pour éviter le catastrophic backtracking, il fait éviter d'avoir deux match possibles pour le même caractère dans un groupe.

## Conclusion

Toute expression régulière peut être compilée automatiquement en un automate non-déterministe, que l'on peut convertir en automate déterministe.

# Prétraitement des textes

## Définitions

- **Jetons**: Nombre d'occurences de mots ($N$)
- **Types**: Nombre de mots distincts dans le vocabulaire ($|V|$)
- **Mot**: Plus petit élément pouvant être prononcé ou écrit en isolation avec un contenu sémantique. On peut accepter les signes de ponctuation dans certaines tâches

Loi de Heaps: $|V| = kN^{\beta}, 0.67 < \beta < 0.75$

Dictionnaires: 

- Oxford English (600 000 mots)
- Petit Robert (60 000 mots)

## Types de traitements

Les traitements les plus fréquents sont:

- Découper le texte en mots (segmentation)
- Uniformiser les mots (normalisation)
- Découper le texte en phrases (segmentation)

## Segmentation de mots (tokenisation)

- Utilisation de délimiteurs (ponctuation et espaces)
  - Difficultés:
	- Apostrophes qui relient des mots
	- Traits d'union reliant deux parties d'une entité ou deux entités séparées
	- Acronymes
	- Mots composés et noms propres
	- Nombres et termes spéciaux
- Choix du tokeniseur:
  - Split
  - Whitespace
  - WordPunct
  - Treebank
  - Tweet
  - Regexp
- Disponible dans toutes les librairies NLP
  - Efficace, utilise des expressions régulières et correction avec un dictionnaire

## Normalisation

Utile en recherche d'information

- On ne veut pas conserver tous les termes
- Classes d'équivalences
  - Suppression de la ponctuation dans les abréviations
  - Remplacement des majuscules par des minuscules
  - Retrait des accords en genre et en nombre
  
Techniques, en ordre décroissant de complexité:

- Analyse morphologique
- Lemmatisation
- Stemming

### Analyse morphologique

- Étude de la composition des mots en morphèmes
- **Morphèmes** : Plus petite unité lexicale ayant un sens spécifique
  - **Racines** : Morphèmes lexicaux, (*stems*)
	- Central; Donne le sens au mot
	- Radical ou lexème
  - **Affixes** : Morphèmes grammaticaux
	- Préfixe (**pré**historique), suffixes (préhistor**ique**) et infixe (t**a**ke, t**oo**k)
	- Deux classes:
		- Dérivatifs: modifie la syntaxe (**-ment** change les adjectifs en adverbes)
		- Flexionnels: pas de dérivation du mot
	- Peuvent être composés

### Lemmatisation

Réduction à une seule forme de base, le lemme

- **Lemme**: Unité lexicale qui porte le sens, pas nécessairement présente dans la composition du mot, contrairement au lexème
- Dictionnaire de formes de base
- Nécessite de connaître la nature des mots (*POS*)

Approches

- Fonction `morphy` de WordNet, utilise des suffixes pour trouver le mot dans le lexique
- Approches probabilistes avec *Hidden Markov Models (HMM)*

### Stemming

Consiste à trouver la racine du mot par troncature et des règles de réécriture.

- Aucun lexique
- Dépend de la langue
  - Anglais: Lovins, Porter, Snowball
  - Français: Snowball
- Peut faire des erreurs

## Segmentation de phrases

Découpage d'un texte en phrases

- Délimitation par un signe de ponctuation

|Signe|Précision|Exemple|
|---|---|---|
|! et ?|Assez fiable, sauf dans les dialogues|"Pourquoi aussi êtes-vous si insatiable?" répondit le renard.|
|.|Quelques ambiguités, abbrébiations et nombres|Monsieur Brossard, ing., Ph.D., est arrivé.|
|...|Interruption ou hésitation|Mais... c'est mon ami !|

- Approximation à l'aide de règles
  - Mot suivant débute par majuscule
  - Mot précédent tout en majuscules
  - Précédé d'une abréviation
  - Précédé et suivi de caractères sans espaces
  - Suivi de plusieurs espaces
  - Précédé ou suivi de signes de ponctuation
  - Suivi de guillements
  - ...
- Une décision pour chaque signe de ponctuation selon le contexte
- Meilleure performance: Apprentissage automatique, classification binaire

## Distance minimale d'édition

Utilisations:

- Distance entre les mots (correction d'orthographe)
- Distance entre des séquences de mots (similarité dans le discours)

Distance de Levenshtein: nombre minimal d'opérations nécessaires pour transformer un mot en un autre.

- Alignement des chaînes de caractères
- Édition par exploration: 
  - Branche la plus courte
  - Problèmes:
	- Nombre exponentiel de chemins
	- Plusieurs chemins pour le même résultat
	- Séquences potentiellement inefficaces
	
On utilise la programmation dynamique:

- Sous-ensemble de caractères
- Conserver les résultats dans un tableau
- On calcule la distance avec ces résultats partiels

$$
\mathit{distance}[i,j]=
\mathit{min}
\begin{cases}
\mathit{distance}[i-1,j]+{C}_{\mathit{ins}}({\mathit{target}}_{i-1}) \\
\mathit{distance}[i-1,j-1]+{C}_{\mathit{subst}}({\mathit{source}}_{j-1},{\mathit{target}}_{i-1}) \\
\mathit{distance}[i,j-1]+{C}_{\mathit{del}}({\mathit{source}}_{j-1})
\end{cases}
$$

Résolution:

- On remplit le tableau de gauche à droite et de bas en haut
- Pour remplir le tableau
  - Si les deux caractères sont différents: on prend le minimum des trois cases adjacentes, et on ajoute le coût de l'opération 
	- horizontal = insertion
	- vertical = suppression
	- diagonale = substitution
  - Si les caractères sont identiques, on reprend la valeur sur la diagonale


# Modèles de langue N-grammes

L'objectif d'un modèle de langue est classifier un texte selon son vocabulaire ou de prédire le prochain mot.

- Observer les mots précédents
- Conserver des comptes relativement simples des mots

## Définition

- **$N$-gramme**: Séquence de $N$ mots
  - Les plus fréquents sont les bigrammes (2) et trigrammes(3)
- **Corpus**:  Ensemble de documents représentatif de notre application ou du sujet d'intérêt

## Applications

La prédiction du mot suivant est utilisée dans plusieurs domaines

- Reconnaissance vocale
- Correction d'orthographe
- Traduction automatique

Classification de textes, désambiguation

## Modèles

Approches:

- Estimer les probabilités à partir des fréquences relatives des mots
- Compter les sous-séquences dans un corpus

Décompte des mots:

- On utilise seulement les formes de mots (les mots tel quels)

Calcul:

- On évalue la probabilité d'une séquence de longueur $N$ à partir de la probabilité de la séquence de longueur $N-1$.
  - Ex: $P(w_4|w_1,w_2,w_3)$
- On évalue la probabilité d'une séquence entière
  - $P(W) = P(w_1,w_2,w_3,w_4)$

Règle de Bayes:

$$
\begin{aligned}
P(B|A) &= \frac{P(A,B)}{P(A)} 
&= \frac{P(B)P(A|B)}{P(A)} 
\end{aligned}
$$

Exemple:

$$
\begin{aligned}
&P(\texttt{nous} | \texttt{les minions sont de retour et}) \\
&=\frac{P(\texttt{les minions sont de retour et nous})}{P(\texttt{les minions sont de retour et})}
\end{aligned}
$$

## Chainage de probabilités

- Selon Google, les comptes de ces séquences sont de 0. 
- Solution: Règle de chaînage de probabilité 
  - Assumer une forme d'indépendance entre les mots.
  - $P(ABC) = P(A) P(B|A) P(C|AB)$
- Forme générale:

$$
\begin{aligned}
P(w_i^n) &= P(w_1)P(w_2|w_1)P(w_3|w_1^2) \ldots P(w_n|w_1^{n-1}) \\
&= \prod_{k=1}^{n} P(w_k|w_1^{k-1})
\end{aligned}
$$

- Exemple:

$$
\begin{aligned}
P(\texttt{les minions sont de retour}) &= \\
P(\texttt{les}) \times \\
P(\texttt{minions}|\texttt{les}) \times \\
P(\texttt{sont}|\texttt{les minions}) \times \\
P(\texttt{de}|\texttt{les minions sont}) \times \\
P(\texttt{retour}|\texttt{les minions sont de}) \times \\
\end{aligned}
$$

Encore une fois, il y a trop de phrases possibles pour pouvoir calculer les statistiques de longs historiques. On pose une autre hypothèse d'indépendance, l'hypothèse de Markov.

$$
\begin{aligned}
P(\texttt{retour}|\texttt{les minions sont de}) 
&\sim P(\texttt{retour}|\texttt{de})
&\sim P(\texttt{retour}|\texttt{sont de})
\end{aligned}
$$

**Hypothèse de Markov**: Les mots sont indépendants du début de l'historique, et seulement des $N$ mots précédents.

$$
P(w_n|w_1{n-1}) \sim P(w_n|w_{n-N+1}^{n-1})
$$

## Estimation des probabilités

Estimateur du maximum de vraisemblance pour les N-grammes, déterminé à partir d'un corpus d'entraînement.

$$
\begin{aligned}
P(w_i|w_{i-1}) &= \frac{\texttt{count}(w_{i-1},w_i)}{\texttt{count}(w_{i-1})} \\
&= \frac{C(w_{i-1},w_i)}{C(w_{i-1})} \\
&= \frac{C(w_{n-1}w_n)}{\sum_{w \in V}C(w_{n-1}w)}
\end{aligned}
$$

Pour calculer les probabilités des bigrammes, on doit faire le compte des unigrammes et des bigrammes, puis, on divise le compte du bigramme par le compte de l'unigramme

Ex: 

$$
\begin{aligned}
P(\texttt{want}|\texttt{I}) \\ 
&= \frac{P(\texttt{I want})}{P(\texttt{I})} \\
&= \frac{827}{2533} = 0.3265
\end{aligned}
$$

- Pour éviter de calculer des probabilités trop petites, on utilise des log-probabilités que l'on additionne au lieu de multiplier.
- On doit délimiter les débuts `<s>` et les fins de phrases `</s>`. On ajoute aussi la fin de phrase au vocabulaire.
- On valide la capacité de généralisation du modèle en générant des phrases ou en analysant les modèles N-grammes (aspects reliés à une langue).

### Génération aléatoire

- Choisir un bigramme aléatoire `(<s>,w)`
- Échantillonner un bigramme `(w,x)`
- Itérer jusqu'à l'obtention d'un bigramme `(x,</s>)`

### Différents corpus

Shakespeare:

- N=884647 jetons
- V=29066 types
- 300000 types de bigrammes, sur 844M possibles, soit 0.04%

### Évaluation

Les textes générés avec des modèles d'ordres supérieurs sont plus cohérents. Ils ressemblent aux textes utilisés pour construire les modèles.

Méthodes de validation intrinsèques:

- Holdout: ensembles d'entraînement et de test
- Métrique d'évaluation: 
  - Perplexité (à minimiser)
- Limites:
  - Données de test peuvent être très différentes des données d'entraînement
  - La perplexité est une mauvaise approximation dans ce cas

$$
\begin{aligned}
PP(W) &= \sqrt[N]{\frac{1}{P(w_1,\ldots,w_N)}} \\
&= \sqrt[N]{\prod_{i=1}^{N}\frac{1}{P(w_i|w_1,\ldots,w_{i-1})}}
\end{aligned}
$$

- Méthodes extrinsèques: tester deux modèles dans une application et prendre le meilleur selon une métrique propre à l'application.

## Lissage

- Les bigrammes qui sont rares ont souvent une probabilité de 0 avec la méthode du maximum de vraisemblance, car ils ne seront pas rencontrés dans le corpus d'entraînement.
- On veut les considérer dans nos calculs: on applique un lissage

### Lissage de Laplace

- Consiste à ajouter un compte de 1 à tous les n-grammes. On augmente donc la taille du corpus de $|V|$ jetons.

La probabilité de rencontrer le mot $w_i$ devient:

$$
P_{\texttt{Laplace}}(w_i) = \frac{c_i+1}{N+V}
$$

Les comptes reconstruits sont:

$$
c_i^{\star} = (c_i+1)\frac{N}{N+V}
$$

Appliqué aux bigrammes, ces équations deviennent:

$$
\begin{aligned}
P_{\texttt{Laplace}}(w_n|w_{n-1}) &= \frac{C(w_{n-1}w_n)+1}{C(w_{n-1})+V} \\
c^{\star}(w_{n-1}w_n) &= (C(w_{n-1}w_n)+1)\frac{C(w_{n-1})}{C(w_{n-1})+V}
\end{aligned}
$$

- Enjeux
  - Approche trop drastique
	- On peut utiliser une valeur $\delta \in [0,1]$, mais ce choix est difficile
  - Pas utilisé en pratique pour les N-grammes, il y a de meilleures méthodes.
  - Utilisé pour lisser d'autres types de modèles probabilistes.
	- Classification de textes
	- Études pilotes
	- Lorsqu'il n'y a pas beaucoup de zéros.

### Autres approches de lissage:
- Si $C(xyz)=0$, on peut utiliser $p(z|y)$ et $p(z)$
- Interpolation (combinaison)
	- Interpolation simple:
	$$
	\begin{aligned}
	\hat{P}(w_n|w_{n-2}w_{n-1}) &= \lambda_{1}P(w_n|w_{n-2}w_{n-1})\\
	&+\lambda_{2}P(w_n|w_{n-1})\\
	&+\lambda_{3}P(w_n)\\
	\sum_{i} \lambda_i&=1
	\end{aligned}
	$$
	- Version plus sophistiquée:
	$$
	\begin{aligned}
	\hat{P}(w_n|w_{n-2}w_{n-1}) &= \lambda_{1}(w_{n-2}^{n-1})P(w_n|w_{n-2}w_{n-1})\\
	&+\lambda_{2}(w_{n-2}^{n-1})P(w_n|w_{n-1})\\
	&+\lambda_{3}(w_{n-2}^{n-1})P(w_n)
	\end{aligned}
	$$
	- Les lambdas sont déterminés à partir d'un jeu de validation (devset) séparé du jeu d'entraînement et de celui de test.
- Backoff (on utilise le premier non-nul en ordre décroissant: $C(zyx)$, $C(zy)$, $C(z)$
  - Katz Backoff
  - Stupid (web) backoff. Cet algorithme ne génère pas une distribution de probabilité, on identifie alors la mesure par $S$.
	$$
	\begin{aligned}
	S(w_i|w_{i-k+1}^{i-1}) &= 
	\begin{cases} 
		\frac{C(w_{i-k+1}^i)}{C(w_{i-k+1}^{i-1})}&,\texttt{si } C(w_{i-k+1}^i) >0\\
		0.4S(w_i|C(w_{i-k+2}^{i-1}))&,\texttt{sinon}
	\end{cases} \\
	S(w_i) &= \frac{C(w_i)}{N}\\
	\end{aligned}
	$$

# Classification de textes

## Naive Bayes

## Régression logistique

## Évaluation

## Métriques

## Validation croisée

# Analyse grammaticale

## Classe de mots

## POS Tags

## Analyseur

## Algorithme de Viterbi

## MEMM

## Choix d'attributs

# Analyse syntaxique

## Groupes syntaxiques

## Grammaire

## Analyseur

## CKY

## Probabiliste

## Analyseur de dépendances

## Chunking

# Recherche d'information

# Wordnet

# Cooccurences

# Premier travail pratique
